@@ -2,12 +2,12 @@
 ;;; doxymacs.el --- ELisp package for making doxygen related stuff easier.
 ;;
 ;;
-;; Copyright (C) 2001, 2002, 2003 Ryan T. Sammartino
+;; Copyright (C) 2001, 2002, 2003, 2005 Ryan T. Sammartino
 ;;
 ;; Author: Ryan T. Sammartino <ryants at shaw dot ca>
 ;;      Kris Verbeeck <kris.verbeeck at advalvas dot be>
 ;; Created: 24/03/2001
-;; Version: 1.5.0
+;; Version: 1.6.0
 ;; Keywords: doxygen documentation
 ;;
 ;; This file is NOT part of GNU Emacs or XEmacs.
@@ -28,7 +28,7 @@
 ;;
 ;; Doxymacs homepage: http://doxymacs.sourceforge.net/
 ;;
-;; $Id: doxymacs.el.in,v 1.9 2003/01/26 01:49:56 ryants Exp $
+;; $Id: doxymacs.el.in,v 1.20 2005/06/04 16:05:30 ryants Exp $
 
 ;; Commentary:
 ;;
@@ -107,14 +107,25 @@
 ;;   - C-c d @ will insert grouping comments around the current region.
 ;;
 ;; Doxymacs has been tested on and works with:
-;; - GNU Emacs 20.7.1, 21.1.1, 21.2.1, 21.2.92.1
-;; - XEmacs 21.1 (patch 14), 21.4 (patches 4-10)
+;; - GNU Emacs 20.7.1, 21.1.1, 21.2.1, 21.2.92.1, 21.3
+;; - XEmacs 21.1 (patch 14), 21.4 (patches 4-17)
 ;;
 ;; If you have success or failure with other version of {X}Emacs, please
 ;; let the authors know.
 
 ;; Change log:
 ;;
+;; 04/06/2005 - version 1.6.0
+;; 14/04/2005 - Use doxymacs-url-exists-p to wrap the various ways of
+;;              checking whether a URL exists.
+;;            - Clean up symbol-near-point hack.
+;; 13/04/2005 - feature request #868413: ability to customize the browser
+;;              doxymacs uses to display documentation.
+;; 12/04/2005 - bug #990123: grouping comments do not work.
+;;            - add some missing doxygen keywords.
+;; 01/04/2005 - patch #1024026: use new font-lock-add-keywords function if
+;;              available.
+;; 31/03/2005 - patch #1102042: handle @param[in] etc. constructs
 ;; 25/01/2003 - remove hard coded version number from comments.
 ;;            - add instructions to avoid byte compiling files.
 ;;            - version 1.5.0
@@ -122,7 +133,7 @@
 ;;            - fix bug #665099: missing @var fontification.
 ;;            - fix bug #665372: @example not fontified properly.
 ;;            - fix fontification for other keywords as well.
-;;            - new customisation variable doxymacs-command-character
+;;            - new customisation variable doxymacs command-character
 ;;              which allows for customisation of the character used
 ;;              to introduce Doxygen commands independent of the
 ;;              current style.
@@ -223,10 +234,9 @@
 (require 'custom)
 (require 'xml-parse)
 (require 'url)
-(require 'w3-cus)
 (require 'tempo)
 
-(defconst doxymacs-version "1.5.0"
+(defconst doxymacs-version "1.6.0"
   "Doxymacs version number")
 
 (defun doxymacs-version ()
@@ -303,6 +313,13 @@ Set to non-nil to use the external XML parser."
   :type 'string
   :group 'doxymacs)
 
+(defcustom doxymacs-browse-url-function
+  'browse-url
+  "*Function to call to launch a browser to display Doxygen documentation.
+This function should take one argument, a string representing the URL to
+display."
+  :type 'function
+  :group 'doxymacs)
 
 (defcustom doxymacs-blank-multiline-comment-template
   nil
@@ -381,24 +398,22 @@ Should be an empty string if comments are terminated by end-of-line."
 		 string)
   :group 'doxymacs)
 
-(defcustom doxymacs-group-begin-comment-template
+(defcustom doxymacs-group-comment-start
   nil
-  "A tempo template to begin `doxymacs-insert-grouping-comments'.
+  "A string to begin a grouping comment (`doxymacs-insert-grouping-comments').
 If nil, then a default template based on the current style as indicated
-by `doxymacs-doxygen-style' will be used.
-
-For help with tempo templates, see http://www.lysator.liu.se/~davidk/elisp/"
-  :type 'list
+by `doxymacs-doxygen-style' will be used."
+  :type '(choice (const :tag "None" nil)
+		 string)
   :group 'doxymacs)
 
-(defcustom doxymacs-group-end-comment-template
+(defcustom doxymacs-group-comment-end
   nil
-  "A tempo template to end `doxymacs-insert-grouping-comments'.
+  "A string to end a grouping comment (`doxymacs-insert-grouping-comments').
 If nil, then a default template based on the current style as indicated
-by `doxymacs-doxygen-style' will be used.
-
-For help with tempo templates, see http://www.lysator.liu.se/~davidk/elisp/"
-  :type 'list
+by `doxymacs-doxygen-style' will be used."
+  :type '(choice (const :tag "None" nil)
+		 string)
   :group 'doxymacs)
 
 ;; End of customisable variables
@@ -507,13 +522,18 @@ Key bindings:
     ;; One shot keywords that take no arguments
     (concat "\\([@\\\\]\\(brief\\|li\\|\\(end\\)?code\\|sa"
 	    "\\|note\\|\\(end\\)?verbatim\\|return\\|arg\\|fn"
-	    "\\|hideinitializer\\|showinitializer\\|interface"
+	    "\\|hideinitializer\\|showinitializer"
+	    ;; FIXME
+	    ;; How do I get & # < > % to work?
+	    ;;"\\|\\\\&\\|\\$\\|\\#\\|<\\|>\\|\\%"
+	    "\\|\\$"
 	    "\\|internal\\|nosubgrouping\\|author\\|date\\|endif"
 	    "\\|invariant\\|post\\|pre\\|remarks\\|since\\|test\\|version"
 	    "\\|\\(end\\)?htmlonly\\|\\(end\\)?latexonly\\|f\\$\\|file"
+	    "\\|\\(end\\)?xmlonly\\|\\(end\\)?manonly\\|property"
 	    "\\|mainpage\\|name\\|overload\\|typedef\\|deprecated\\|par"
 	    "\\|addindex\\|line\\|skip\\|skipline\\|until\\|see"
-	    "\\|endlink\\)\\)\\>")
+	    "\\|endlink\\|callgraph\\|endcond\\|else\\)\\)\\>")
     '(0 font-lock-keyword-face prepend))
    ;; attention, warning, etc. given a different font
    (list
@@ -521,14 +541,15 @@ Key bindings:
     '(0 font-lock-warning-face prepend))
    ;; keywords that take a variable name as an argument
    (list
-    (concat "\\([@\\\\]\\(var\\|param\\|a\\|if\\|namespace\\|relates"
-	    "\\|def\\)\\)\\s-+\\(\\sw+\\)")
+    (concat "\\([@\\\\]\\(param\\(?:\\[\\(?:in\\|out\\|in,out\\)\\]\\)?"
+	    "\\|a\\|namespace\\|relates\\(also\\)?"
+	    "\\|var\\|def\\)\\)\\s-+\\(\\sw+\\)")
     '(1 font-lock-keyword-face prepend)
-    '(3 font-lock-variable-name-face prepend))
+    '(4 font-lock-variable-name-face prepend))
    ;; keywords that take a type name as an argument
    (list
-    (concat "\\([@\\\\]\\(class\\|struct\\|union\\|exception"
-	    "\\|throw\\)\\)\\s-+\\(\\sw+\\)")
+    (concat "\\([@\\\\]\\(class\\|struct\\|union\\|exception\\|enum"
+	    "\\|throw\\|interface\\|protocol\\)\\)\\s-+\\(\\(\\sw\\|:\\)+\\)")
     '(1 font-lock-keyword-face prepend)
     '(3 font-lock-type-face prepend))
    ;; keywords that take a function name as an argument
@@ -558,12 +579,24 @@ Key bindings:
     '(2 font-lock-string-face prepend))
    ;; one argument that can contain arbitrary non-whitespace stuff
    (list
-    "\\([@\\\\]\\(link\\|copydoc\\)\\)\\s-+\\([^ \t\n]+\\)"
+    (concat "\\([@\\\\]\\(link\\|copydoc\\|xrefitem"
+	    "\\|if\\(not\\)?\\|elseif\\)\\)"
+	    "\\s-+\\([^ \t\n]+\\)")
     '(1 font-lock-keyword-face prepend)
-    '(3 font-lock-string-face prepend))
+    '(4 font-lock-string-face prepend))
+   ;; one optional argument that can contain arbitrary non-whitespace stuff
+   (list
+    "\\([@\\\\]\\(cond\\|dir\\)\\(\\s-+[^ \t\n]+\\)?\\)"
+    '(1 font-lock-keyword-face prepend)
+    '(3 font-lock-string-face prepend t))
+   ;; one optional argument with no space between
+   (list
+    "\\([@\\\\]\\(~\\)\\([^ \t\n]+\\)?\\)"
+    '(1 font-lock-keyword-face prepend)
+    '(3 font-lock-string-face prepend t))
    ;; one argument that has to be a filename
    (list
-    (concat "\\([@\\\\]\\(example\\|\\(dont\\)?include"
+    (concat "\\([@\\\\]\\(example\\|\\(dont\\)?include\\|includelineno"
 	    "\\|htmlinclude\\|verbinclude\\)\\)\\s-+"
 	    "\\(\"?[~:\\/a-zA-Z0-9_. ]+\"?\\)")
     '(1 font-lock-keyword-face prepend)
@@ -595,10 +628,14 @@ Key bindings:
   "Turn on font-lock for Doxygen keywords."
   ;; FIXME How do I turn *off* font-lock for Doxygen keywords?
   (interactive)
-  (let ((old (if (eq (car-safe font-lock-keywords) t)
+  (if (functionp 'font-lock-add-keywords)
+      ;; Use new (proper?) font-lock-add-keywords function
+      (font-lock-add-keywords nil doxymacs-doxygen-keywords)
+    ;; Use old-school way
+    (let ((old (if (eq (car-safe font-lock-keywords) t)
 		 (cdr font-lock-keywords)
 	       font-lock-keywords)))
-    (setq font-lock-keywords (append old doxymacs-doxygen-keywords))))
+      (setq font-lock-keywords (append old doxymacs-doxygen-keywords)))))
 
 
 
@@ -673,6 +710,39 @@ buffer."
 to comp-list."
   (doxymacs-set-dir-element dir doxymacs-completion-lists comp-list))
 
+(defun doxymacs-url-exists-p (url)
+  "Return t iff the URL exists."
+  (let* ((urlobj (url-generic-parse-url url))
+	 (type (url-type urlobj))
+	 (exists nil))
+    (cond
+     ((equal type "http")
+      (cond
+       ;; Try url-file-exists, if it exists
+       ((fboundp 'url-file-exists)
+	(setq exists (url-file-exists url)))
+       ;; Otherwise, try url-file-exists-p (newer url.el)
+       ((fboundp 'url-file-exists-p)
+	(setq exists (url-file-exists-p url)))
+       ;; Otherwise, try wget
+       ((executable-find (if (eq system-type 'windows-nt) "wget.exe" "wget"))
+	(if (string-match "200 OK"
+			  (shell-command-to-string
+			   (concat "wget -S --spider " url)))
+	    (setq exists t)))
+       ;; Otherwise, try lynx
+       ((executable-find (if (eq system-type 'windows-nt) "lynx.exe" "lynx"))
+	(if (string-match "200 OK"
+			  (shell-command-to-string
+			   (concat "lynx -head -source " url)))
+	    (setq exists t)))
+       ;; Give up.
+       (t (error "Could not find url-file-exists, url-file-exists-p, wget or lynx"))))
+     ((equal type "file")
+      (setq exists (file-exists-p (url-filename urlobj))))
+     (t (error (concat "Scheme " type " not supported for URL " url))))
+    exists))
+
 (defun doxymacs-load-tags (f)
   "Loads a Doxygen generated XML tags file into the buffer *doxytags*."
   (let* ((tags-buffer (doxymacs-filename-to-buffer f))
@@ -698,7 +768,7 @@ to comp-list."
 		      (insert-file-contents xml))
 		  ;; Otherwise, try and grab it as a URL
 		  (progn
-		    (if (url-file-exists xml)
+		    (if (doxymacs-url-exists-p xml)
 			(progn
 			  (set-buffer new-buffer)
 			  (url-insert-file-contents xml)
@@ -860,31 +930,31 @@ completion list."
 
 (defun doxymacs-display-url (root url)
   "Displays the given match."
-  (browse-url (concat root "/" url)))
+  (apply doxymacs-browse-url-function (list (concat root "/" url))))
 
-;; GNU Emacs doesn't have symbol-near-point apparently
+;; Some versions of GNU Emacs don't have symbol-near-point apparently
 ;; stolen from browse-cltl2.el, and in turn:
 ;; stolen from XEmacs 19.15 syntax.el
-(if (not (fboundp (function symbol-near-point)))
-    (defun symbol-near-point ()
-      "Return the first textual item to the nearest point."
-      (interactive)
-      ;;alg stolen from etag.el
-      (save-excursion
-	(if (not (memq (char-syntax (preceding-char)) '(?w ?_)))
-	    (while (not (looking-at "\\sw\\|\\s_\\|\\'"))
-	      (forward-char 1)))
-	(while (looking-at "\\sw\\|\\s_")
-	  (forward-char 1))
-	(if (re-search-backward "\\sw\\|\\s_" nil t)
-	    (regexp-quote
-	     (progn (forward-char 1)
-		    (buffer-substring (point)
-				      (progn (forward-sexp -1)
-					     (while (looking-at "\\s'")
-					       (forward-char 1))
-					     (point)))))
-	  nil))))
+(defun doxymacs-symbol-near-point ()
+  "Return the first textual item to the nearest point."
+  (if (fboundp 'symbol-near-point)
+      (symbol-near-point)
+    ;;alg stolen from etag.el
+    (save-excursion
+      (if (not (memq (char-syntax (preceding-char)) '(?w ?_)))
+	  (while (not (looking-at "\\sw\\|\\s_\\|\\'"))
+	    (forward-char 1)))
+      (while (looking-at "\\sw\\|\\s_")
+	(forward-char 1))
+      (if (re-search-backward "\\sw\\|\\s_" nil t)
+	  (regexp-quote
+	   (progn (forward-char 1)
+		  (buffer-substring (point)
+				    (progn (forward-sexp -1)
+					   (while (looking-at "\\s'")
+					     (forward-char 1))
+					   (point)))))
+	nil))))
 
 (defun doxymacs-lookup (symbol &optional filename)
   "Look up the symbol under the cursor in Doxygen generated documentation."
@@ -903,7 +973,7 @@ completion list."
 	   (let ((symbol (completing-read
 			  "Look up: "
 			  completion-list nil nil
-			  (symbol-near-point)))
+			  (doxymacs-symbol-near-point)))
 		 (filename f))
 	     (list symbol filename)))))))
   (let ((url (doxymacs-symbol-completion
@@ -1167,36 +1237,6 @@ the completion or nil if canceled by the user."
 	 nil))))
  "Default C++-style template for function documentation.")
 
-(defconst doxymacs-JavaDoc-group-begin-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(" //@{" > n)
-  "Default JavaDoc-style template for beginning-of-group comment.")
-
-(defconst doxymacs-JavaDoc-group-end-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(n " //@}" >)
-  "Default JavaDoc-style template for end-of-group comment.")
-
-(defconst doxymacs-Qt-group-begin-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(" /*@{*/" > n)
-  "Default Qt-style template for beginning-of-group comment.")
-
-(defconst doxymacs-Qt-group-end-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(n " /*@}*/" >)
-  "Default Qt-style template for end-of-group comment.")
-
-(defconst doxymacs-C++-group-begin-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(" /// @{" > n)
-  "Default C++-style template for beginning-of-group comment.")
-
-(defconst doxymacs-C++-group-end-comment-template
-  ;; The leading space is a hack to get the indentation to work properly
-  '(n " /// @}" >)
-  "Default C++-style template for end-of-group comment.")
-
 (defun doxymacs-invalid-style ()
   "Warn the user that he has set `doxymacs-doxygen-style' to an invalid
 style."
@@ -1343,13 +1383,34 @@ the column given by `comment-column' (much like \\[indent-for-comment])."
 (defun doxymacs-insert-grouping-comments (start end)
   "Inserts doxygen grouping comments around the current region."
   (interactive "*r")
-  (save-excursion
-    (goto-char start)
-    (beginning-of-line)
-    (doxymacs-call-template "group-begin-comment")
-    (goto-char end)
-    (end-of-line)
-    (doxymacs-call-template "group-end-comment")))
+  (let* ((starter  (or doxymacs-group-comment-start
+		      (cond
+		       ((string= doxymacs-doxygen-style "JavaDoc")
+			"//@{")
+		       ((string= doxymacs-doxygen-style "Qt")
+			"/*@{*/")
+		       ((string= doxymacs-doxygen-style "C++")
+			"/// @{")
+		       (t
+			(doxymacs-invalid-style)))))
+	 (ender (or doxymacs-group-comment-end
+		    (cond
+		       ((string= doxymacs-doxygen-style "JavaDoc")
+			"//@}")
+		       ((string= doxymacs-doxygen-style "Qt")
+			"/*@}*/")
+		       ((string= doxymacs-doxygen-style "C++")
+			"/// @}")
+		       (t
+			(doxymacs-invalid-style))))))
+    (save-excursion
+      (goto-char end)
+      (end-of-line)
+      (insert ender)
+      (goto-char start)
+      (beginning-of-line)
+      (insert starter))))
+
 
 
 ;; These are helper functions that search for the next function
